"""Declarative ORM event decorators and event registration.

SQLAlchemy features an ORM event API but one thing that is lacking is a way to
register event handlers in a declarative way inside the Model's class
definition. To bridge this gap, this module contains a collection of decorators
that enable this kind of functionality.

Instead of having to write event registration like this::

    from sqlalchemy import event

    from project.core import Model

    class User(Model):
        _id = Column(types.Integer(), primary_key=True)
        email = Column(types.String())

    def set_email_listener(target, value, oldvalue, initiator):
        print 'received "set" event for target: {0}'.format(target)
        return value

    def before_insert_listener(mapper, connection, target):
        print 'received "before_insert" event for target: {0}'.format(target)

    event.listen(User.email, 'set', set_email_listener, retval=True)
    event.listen(User, 'before_insert', before_insert_listener)

Model Events allows one to write event registration more succinctly as::

    from alchy import events

    from project.core import Model

    class User(Model):
        _id = Column(types.Integer(), primary_key=True)
        email = Column(types.String())

        @events.on_set('email', retval=True)
        def on_set_email(target, value, oldvalue, initiator):
            print 'received set event for target: {0}'.format(target)
            return value

        @events.before_insert()
        def before_insert(mapper, connection, target):
            print ('received "before_insert" event for target: {0}'
                   .format(target))

For details on each event type's expected function signature, see
`SQLAlchemy's ORM Events
<http://docs.sqlalchemy.org/en/latest/orm/events.html>`_.
"""
# pylint: disable=invalid-name

import sqlalchemy

from ._compat import iteritems


__all__ = [
    'on_set',
    'on_append',
    'on_remove',
    'before_delete',
    'before_insert',
    'before_update',
    'before_insert_update',
    'after_delete',
    'after_insert',
    'after_update',
    'after_insert_update',
    'on_append_result',
    'on_create_instance',
    'on_instrument_class',
    'before_configured',
    'after_configured',
    'on_mapper_configured',
    'on_populate_instance',
    'on_translate_row',
    'on_expire',
    'on_load',
    'on_refresh'
]


class Event(object):
    """Universal event class used when registering events."""
    def __init__(self, name, attribute, listener, kargs):
        self.name = name
        self.attribute = attribute
        self.listener = listener
        self.kargs = kargs


class GenericEvent(object):
    """Base class for generic event decorators."""
    event_names = None

    def __init__(self, **event_kargs):
        self.attribute = None
        self.event_kargs = event_kargs

    def __call__(self, func):
        return make_event(self.event_names,
                          self.attribute,
                          **self.event_kargs)(func)


class AttributeEvent(GenericEvent):
    """Base class for an attribute event decorators."""
    def __init__(self, attribute, **event_kargs):
        self.attribute = attribute
        self.event_kargs = event_kargs


##
# Attribute Events
# http://docs.sqlalchemy.org/en/latest/orm/events.html#attribute-events
##


class on_set(AttributeEvent):
    """Event decorator for the ``set`` event."""
    event_names = 'set'


class on_append(AttributeEvent):
    """Event decorator for the ``append`` event."""
    event_names = 'append'


class on_remove(AttributeEvent):
    """Event decorator for the ``remove`` event."""
    event_names = 'remove'


##
# Mapper Events
# http://docs.sqlalchemy.org/en/latest/orm/events.html#mapper-events
##


class before_delete(GenericEvent):
    """Event decorator for the ``before_delete`` event."""
    event_names = 'before_delete'


class before_insert(GenericEvent):
    """Event decorator for the ``before_insert`` event."""
    event_names = 'before_insert'


class before_update(GenericEvent):
    """Event decorator for the ``before_update`` event."""
    event_names = 'before_update'


class before_insert_update(GenericEvent):
    """Event decorator for the ``before_insert`` and ``before_update`` events.
    """
    event_names = ['before_insert', 'before_update']


class after_delete(GenericEvent):
    """Event decorator for the ``after_delete`` event."""
    event_names = 'after_delete'


class after_insert(GenericEvent):
    """Event decorator for the ``after_insert`` event."""
    event_names = 'after_insert'


class after_update(GenericEvent):
    """Event decorator for the ``after_update`` event."""
    event_names = 'after_update'


class after_insert_update(GenericEvent):
    """Event decorator for the ``after_insert`` and ``after_update`` events."""
    event_names = ['after_insert', 'after_update']


class on_append_result(GenericEvent):
    """Event decorator for the ``append_result`` event."""
    event_names = 'append_result'


class on_create_instance(GenericEvent):
    """Event decorator for the ``create_instance`` event."""
    event_names = 'create_instance'


class on_instrument_class(GenericEvent):
    """Event decorator for the ``instrument_class`` event."""
    event_names = 'instrument_class'


class before_configured(GenericEvent):
    """Event decorator for the ``before_configured`` event."""
    event_names = 'before_configured'


class after_configured(GenericEvent):
    """Event decorator for the ``after_configured`` event."""
    event_names = 'after_configured'


class on_mapper_configured(GenericEvent):
    """Event decorator for the ``mapper_configured`` event."""
    event_names = 'mapper_configured'


class on_populate_instance(GenericEvent):
    """Event decorator for the ``populate_instance`` event."""
    event_names = 'populate_instance'


class on_translate_row(GenericEvent):
    """Event decorator for the ``translate_row`` event."""
    event_names = 'translate_row'


##
# Instance Events
# http://docs.sqlalchemy.org/en/latest/orm/events.html#instance-events
##


class on_expire(GenericEvent):
    """Event decorator for the ``expire`` event."""
    event_names = 'expire'


class on_load(GenericEvent):
    """Event decorator for the ``load`` event."""
    event_names = 'load'


class on_refresh(GenericEvent):
    """Event decorator for the ``refresh`` event."""
    event_names = 'refresh'


def register(cls, dct):
    """Register events defined on a class during metaclass creation."""

    events = []

    # append class attribute defined events
    if dct.get('__events__'):
        # Events defined on __events__ can have many forms (e.g. string based,
        # list of tuples, etc). So we need to iterate over them and parse into
        # standardized Event object.
        for event_name, listeners in iteritems(dct['__events__']):
            if not isinstance(listeners, list):
                listeners = [listeners]

            for listener in listeners:
                if isinstance(listener, tuple):
                    # listener definition includes event.listen keyword args
                    listener, kargs = listener
                else:
                    kargs = {}

                if not callable(listener):
                    # assume listener is a string reference to class method
                    listener = dct[listener]

                events.append(Event(event_name,
                                    kargs.pop('attribute', None),
                                    listener,
                                    kargs))

    # add events which were added via @event decorator
    for value in dct.values():
        if hasattr(value, '__event__'):
            if not isinstance(value.__event__, list):  # pragma: no cover
                value.__event__ = [value.__event__]
            events.extend(value.__event__)

    if events:
        # Reassemble events dict into consistent form using Event objects as
        # values.
        events_dict = {}
        for evt in events:
            if evt.attribute is None:
                obj = cls
            else:
                obj = getattr(cls, evt.attribute)

            if evt.name.startswith('on_'):
                event_name = evt.name.replace('on_', '', 1)
            else:
                event_name = evt.name

            sqlalchemy.event.listen(obj, event_name, evt.listener, **evt.kargs)
            events_dict.setdefault(evt.name, []).append(evt)

        dct['__events__'].update(events_dict)


def make_event(event_names, attribute=None, **event_kargs):
    """Generic event decorator maker which attaches metadata to function object
    so that :func:`register` can find the event definition.
    """
    def decorator(func):
        """Function decorator that attaches an `__event__` attribute hook which
        is expected when registering a method as an event handler. See
        :func:`register` for details on how this is implemented.
        """
        if not hasattr(func, '__event__'):
            # Set initial value to list so function can handle multiple events.
            func.__event__ = []

        # NOTE: Have to assign to a separate variable name due to global name
        # access issues.
        if not isinstance(event_names, (list, tuple)):
            _event_names = [event_names]
        else:
            _event_names = event_names

        # Attach event object to function which will be picked up in
        # `register()`.
        func.__event__ += [Event(event_name, attribute, func, event_kargs)
                           for event_name in _event_names]

        # Return function as-is since method definition should be compatible
        # with sqlalchemy.event.listen().
        return func
    return decorator



"""Declarative base for ORM models.
"""

from sqlalchemy import inspect, orm
from sqlalchemy.ext.declarative import (
    declarative_base,
    DeclarativeMeta,
    declared_attr,
)

from . import query, events
from .utils import (
    is_sequence,
    camelcase_to_underscore,
    get_mapper_class,
    merge_declarative_args,
    should_set_tablename,
)
from ._compat import iteritems


__all__ = [
    'ModelMeta',
    'ModelBase',
    'make_declarative_base',
    'extend_declarative_base',
]


class ModelMeta(DeclarativeMeta):
    """ModelBase's metaclass which provides:

    - Tablename autogeneration when ``__tablename__`` not defined.
    - Support for multiple database bindings via \
    :attr:`ModelBase.__bind_key__`.
    - Support for declarative ORM events via :mod:`alchy.events` decorators \
    or :attr:`ModelBase.__events__`.
    """
    def __new__(mcs, name, bases, dct):
        # Set __events__ to expected default so that it's updatable when
        # initializing. E.g. if class definition sets __events__=None but
        # defines decorated events, then we want the final __events__ attribute
        # to reflect the registered events. If set to anything that's
        # non-empty/non-dict will lead to an error if decorated events defined.
        if not dct.get('__events__'):
            dct['__events__'] = {}

        if '__bind_key__' not in dct:
            base_dcts = [dct] + [base.__dict__ for base in bases]
            for base in base_dcts:
                if '__bind_key__' in base:
                    dct['__bind_key__'] = base['__bind_key__']
                    break

        cls = DeclarativeMeta.__new__(mcs, name, bases, dct)

        # Determine if should set __tablename__.
        # This is being done after DeclarativeMeta.__new__()
        # as the class is needed to accommodate @declared_attr columns.
        if should_set_tablename(cls):
            # Set to underscore version of class name.
            cls.__tablename__ = camelcase_to_underscore(name)

        return cls

    def __init__(cls, name, bases, dct):
        DeclarativeMeta.__init__(cls, name, bases, dct)

        if hasattr(cls, '__table__'):
            if '__bind_key__' in dct:
                cls.__table__.info['bind_key'] = dct['__bind_key__']

            events.register(cls, dct)


class ModelBase(object):
    """Base class for creating a declarative base for models.

    To create a declarative base::

        # in project/core.py
        from alchy import ModelBase, make_declarative_base

        class Base(ModelBase):
            # augument the ModelBase with super powers
            pass

        Model = make_declarative_base(Base=Base)


        # in project/models/user.py
        from project.core import Model

        class User(Model):
            # define declarative User model
            pass


    Attributes:

        __table_args__: Default table args.

        __mapper_args__: Define a default order by when not specified by query
            operation, e.g.: ``{ 'order_by': [column1, column2] }``

        __bind_key__: Bind a model to a particular database URI using keys from
            ``Manager.config['SQLALCHEMY_BINDS']``. By default a model will be
            bound to ``Manager.config['SQLALCHEMY_DATABASE_URI']``.

        __events__: Register orm event listeners. See :mod:`alchy.events` for
            more details.

        query_class: Query class to use for ``self.query``.

        query: An instance of :attr:`query_class`. Can be used to query the
            database for instances of this model. NOTE: Requires setting
            ``MyClass.query = QueryProperty(session)`` when session available.
            See :func:`make_declarative_base` for automatic implementation.
    """

    __bind_key__ = None
    __events__ = {}
    query_class = query.QueryModel
    query = None

    @declared_attr
    def __mapper_args__(cls):  # pylint: disable=no-self-argument
        _, kargs = merge_declarative_args(cls,
                                          '__global_mapper_args__',
                                          '__local_mapper_args__')
        return kargs

    @declared_attr
    def __table_args__(cls):  # pylint: disable=no-self-argument
        args, kargs = merge_declarative_args(cls,
                                             '__global_table_args__',
                                             '__local_table_args__')
        # Append kargs onto end of args to adhere to SQLAlchemy requirements.
        args.append(kargs)
        return tuple(args)

    def __init__(self, *args, **kargs):
        """Initialize model instance by calling :meth:`update`."""
        self.update(*args, **kargs)

    def __repr__(self):  # pragma: no cover
        values = ', '.join(['{0}={1}'.format(c, repr(getattr(self, c)))
                            for c in self.columns()])
        return '<{0}({1})>'.format(self.__class__.__name__, values)

    def __getitem__(self, item):
        """Proxy getitem to getattr."""
        return getattr(self, item)

    def __setitem__(self, item, value):
        """Proxy setitem to setattr."""
        setattr(self, item, value)

    def update(self, data_dict=None, **kargs):
        """Update model with arbitrary set of data."""

        data = data_dict if isinstance(data_dict, dict) else kargs

        for field, value in iteritems(data):
            if hasattr(self, field):
                self._set_field(field, value)

    def _set_field(self, field, value):
        """Set model field with value."""
        # Consider value a dict if any of its elements are a dict.
        if is_sequence(value):
            is_dict = any([isinstance(val, dict) for val in value])
        else:
            is_dict = isinstance(value, dict)

        attr = getattr(self, field)

        if (hasattr(attr, 'update') and
                value and
                is_dict and
                not isinstance(attr, dict)):
            # Nest calls to attr.update.
            attr.update(value)
        else:
            if field in self.relationships():
                self._set_relationship_field(field, value)
            else:
                setattr(self, field, value)

    def _set_relationship_field(self, field, value):
        """Set model relationships field with value."""
        relationship_class = get_mapper_class(self.__class__, field)
        is_sequence_relationship = is_sequence(getattr(self, field))

        if is_sequence_relationship and is_sequence(value):
            # Convert each value instance to relationship class.
            value = [relationship_class(val) if isinstance(val, dict)
                     else val
                     for val in value]
        elif not value and isinstance(value, dict):
            # If value is {} and we're trying to update a relationship
            # attribute, then we need to set to None to nullify relationship
            # value.
            value = None
        elif not is_sequence_relationship and isinstance(value, dict):
            # Convert single value to relationship class.
            value = relationship_class(value)

        setattr(self, field, value)

    @property
    def __to_dict__(self):
        """Configuration for :meth:`to_dict`. Do any necessary preprocessing
        and return a set of string attributes which represent the fields which
        should be returned when calling :meth:`to_dict`.

        By default this model is refreshed if it's ``__dict__`` state is empty
        and only the ORM descriptor fields are returned.

        This is the property to override if you want to return more/less than
        the default ORM descriptor fields.

        Generally, we can usually rely on ``__dict__`` as a representation of
        model when it's just been loaded from the database. In this case,
        whatever values are present in ``__dict__`` are the loaded values
        from the database which include/exclude lazy attributes (columns and
        relationships).

        One issue to be aware of is that after a model has been committed (or
        expired), ``__dict__`` will be empty. This can be worked around by
        calling :meth:`refresh` which will reload the data from the database
        using the default loader strategies.

        These are the two main cases this default implementation will try to
        cover. For anything more complex it would be best to override this
        property or the :meth:`to_dict` method itself.
        """
        if not self.descriptor_dict.keys() and orm.object_session(self):
            # If the descriptor dict keys are empty, assume we need to refresh.
            self.refresh()

        return set(self.descriptor_dict.keys())

    @property
    def descriptor_dict(self):
        """Return :attr:`__dict__` key-filtered by :attr:`descriptors`."""
        return dict([(key, value)
                     for key, value in iteritems(self.__dict__)
                     if key in self.descriptors()])

    def to_dict(self):
        """Return dict representation of model by filtering fields using
        :attr:`__to_dict__`.
        """
        data = {}
        relationships = self.relationships()

        for field in self.__to_dict__:
            value = getattr(self, field)

            # Nest calls to `to_dict`. Try to find method on base value,
            # sequence values, or dict values.
            if hasattr(value, 'to_dict'):
                value = value.to_dict()
            elif is_sequence(value):
                value = [v.to_dict() if hasattr(v, 'to_dict') else v
                         for v in value]
            elif isinstance(value, dict):
                value = dict([(k, v.to_dict() if hasattr(v, 'to_dict') else v)
                              for k, v in iteritems(value)])
            elif field in relationships and value is None:
                value = {}

            data[field] = value

        return data

    def __iter__(self):
        """Implement ``__iter__`` so model can be converted to dict via
        ``dict()``.
        """
        return iteritems(self.to_dict())

    ##
    # session based methods/properties
    ##

    @property
    def object_session(self):
        """Return session belonging to ``self``"""
        return orm.object_session(self)

    @classmethod
    def session(cls):
        """Return session from query property"""
        return cls.query.session

    def flush(self, *args, **kargs):
        """Call ``session.flush()`` on ``self``"""
        self.session().flush([self], *args, **kargs)

    def save(self, *args, **kargs):
        """Call ``session.add()`` on ``self``"""
        self.session().add(self, *args, **kargs)

    def delete(self, *args, **kargs):
        """Call ``session.delete()`` on ``self``"""
        return self.session().delete(self, *args, **kargs)

    def expire(self, *args, **kargs):
        """Call ``session.expire()`` on ``self``"""
        self.session().expire(self, *args, **kargs)

    def refresh(self, *args, **kargs):
        """Call ``session.refresh()`` on ``self``"""
        self.session().refresh(self, *args, **kargs)

    def expunge(self, *args, **kargs):
        """Call ``session.expunge()`` on ``self``"""
        self.session().expunge(self, *args, **kargs)

    def is_modified(self, *args, **kargs):
        """Call ``session.is_modified()`` on ``self``"""
        return self.session().is_modified(self, *args, **kargs)

    @classmethod
    def get(cls, *args, **kargs):
        """Proxy to ``cls.query.get()``"""
        return cls.query.get(*args, **kargs)

    @classmethod
    def get_by(cls, data_dict=None, **kargs):
        """Return first instance filtered by values using
        ``cls.query.filter_by()``.
        """
        data = data_dict if isinstance(data_dict, dict) else kargs
        return cls.query.filter_by(**data).first()

    @classmethod
    def filter(cls, *args, **kargs):
        """Proxy to ``cls.query.filter()``"""
        return cls.query.filter(*args, **kargs)

    @classmethod
    def filter_by(cls, *args, **kargs):
        """Proxy to ``cls.query.filter_by()``"""
        return cls.query.filter_by(*args, **kargs)

    ##
    # SQLAlchemy.inspect() based methods/properties
    ##

    @classmethod
    def primary_key(cls):
        """Return primary key as either single column (one primary key) or
        tuple otherwise.
        """
        primary = inspect(cls).primary_key

        if len(primary) == 1:
            primary = primary[0]

        return primary

    @classmethod
    def primary_keys(cls):
        """Return primary keys as tuple."""
        return inspect(cls).primary_key

    @classmethod
    def primary_attrs(cls):
        """Return class attributes from primary keys."""
        primary_keys = cls.primary_keys()
        return [getattr(cls, attr)
                for attr in cls.columns()
                if getattr(cls, attr).property.columns[0] in primary_keys]

    @classmethod
    def attrs(cls):
        """Return ORM attributes"""
        # pylint: disable=no-member
        return inspect(cls).attrs.keys()

    @classmethod
    def descriptors(cls):
        """Return all ORM descriptors"""
        # pylint: disable=maybe-no-member
        return [descr for descr in inspect(cls).all_orm_descriptors.keys()
                if not descr.startswith('__')]

    @classmethod
    def relationships(cls):
        """Return ORM relationships"""
        # pylint: disable=no-member
        return inspect(cls).relationships.keys()

    @classmethod
    def column_attrs(cls):
        """Return table columns as list of class attributes at the class level.
        """
        return inspect(cls).column_attrs

    @classmethod
    def columns(cls):
        """Return table columns."""
        # pylint: disable=no-member
        return inspect(cls).columns.keys()


def make_declarative_base(session=None,
                          Model=None,
                          Base=None,
                          Meta=None,
                          metadata=None):
    """Factory function for either creating a new declarative base class or
    extending a previously defined one.
    """
    if Model is None:
        Base = Base or ModelBase
        Meta = Meta or ModelMeta
        Model = declarative_base(cls=Base,
                                 constructor=Base.__init__,
                                 metaclass=Meta,
                                 metadata=metadata)

    extend_declarative_base(Model, session)

    return Model


def extend_declarative_base(Model, session=None):
    """Extend a declarative base class with additional properties.

    - Extend `Model` with query property accessor
    """
    # Attach query attribute to Model if `session` object passed in.
    if session:
        Model.query = query.QueryProperty(session)



"""Collection of custom column types.
"""

from sqlalchemy.types import SchemaType, TypeDecorator, Enum

from .utils import camelcase_to_underscore
from ._compat import with_metaclass


__all__ = [
    'DeclarativeEnumType',
    'DeclarativeEnum'
]


##
# Declarative Enum type
# Adapted from the Enum Recipe:
# http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/
##

class EnumSymbol(object):
    """Define a fixed symbol tied to a parent class."""

    def __init__(self, enum_class, name, value, description):
        self.enum_class = enum_class
        self.name = name
        self.value = value
        self.description = description

    def __reduce__(self):
        """Allow unpickling to return the symbol linked to the DeclarativeEnum
        class.
        """
        return getattr, (self.enum_class, self.name)

    def __iter__(self):
        return iter([self.value, self.description])

    def __repr__(self):
        return "<%s>" % self.name

    def __str__(self):
        return self.name

    def to_dict(self):
        """Represent symbol as dict."""
        return {'value': self.value, 'description': self.description}


class EnumMeta(type):
    """Generate new DeclarativeEnum classes."""

    def __init__(cls, classname, bases, dct):
        cls._reg = cls._reg.copy()
        for name, value in dct.items():
            if isinstance(value, tuple):
                sym = cls._reg[value[0]] = EnumSymbol(cls, name, *value)
                setattr(cls, name, sym)
        type.__init__(cls, classname, bases, dct)

    def __iter__(cls):
        return iter(cls._reg.values())


class DeclarativeEnumType(SchemaType, TypeDecorator):
    """Column type usable in table column definitions."""

    def __init__(self, enum, name=None):
        self.enum = enum
        enum_args = enum.__enum_args__.copy()

        if name is not None:
            enum_args['name'] = name
        elif 'name' not in enum_args:
            enum_args['name'] = 'ck_' + camelcase_to_underscore(enum.__name__)

        self.impl = Enum(*enum.values(), **enum_args)

    def _set_table(self, table, column):
        self.impl._set_table(table, column)

    def copy(self):
        return DeclarativeEnumType(self.enum)

    def process_bind_param(self, value, dialect):
        if value is None:  # pragma: no cover
            return None
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:  # pragma: no cover
            return None
        return self.enum.from_string(value.strip())


class DeclarativeEnum(with_metaclass(EnumMeta, object)):
    """Declarative enumeration.

    For example::

        class OrderStatus(DeclarativeEnum):
            pending = ('p', 'Pending')
            submitted = ('s', 'Submitted')
            complete = ('c', 'Complete')

        class Order(Model):
            status = Column(OrderStatus.db_type(), default=OrderStatus.pending)
    """

    _reg = {}

    __enum_args__ = {}

    @classmethod
    def from_string(cls, string):
        """Return enum symbol given string value.

        Raises:
            ValueError: If `string` doesn't correspond to an enum value.
        """
        try:
            return cls._reg[string]
        except KeyError:
            raise ValueError('Invalid value for %r: %r' % (cls.__name__,
                                                           string))

    @classmethod
    def values(cls):
        """Return list of possible enum values. Each value is a valid argument
        to :meth:`from_string`.
        """
        return cls._reg.keys()

    @classmethod
    def db_type(cls, name=None):
        """Return database column type for use in table column definitions."""
        return DeclarativeEnumType(cls, name=name)



"""Generic utility functions used in package.

..
    >>> from alchy.utils import *
"""

import re
import warnings
from collections import Iterable

from sqlalchemy import Column
from sqlalchemy.exc import SAWarning
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr

from ._compat import string_types, itervalues, iteritems, classmethod_func


__all__ = [
    'is_sequence',
    'camelcase_to_underscore',
    'iterflatten',
    'flatten'
]


def is_sequence(obj):
    """Test if `obj` is an iterable but not ``dict`` or ``str``. Mainly used to
    determine if `obj` can be treated like a ``list`` for iteration purposes.
    """
    return (isinstance(obj, Iterable) and
            not isinstance(obj, string_types) and
            not isinstance(obj, dict))


def camelcase_to_underscore(string):
    """Convert string from ``CamelCase`` to ``under_score``."""
    return re.sub('((?<=[a-z0-9])[A-Z]|(?<!_)(?!^)[A-Z](?=[a-z]))', r'_\1',
                  string).lower()


def iterflatten(items):
    """Return iterator which flattens list/tuple of lists/tuples::

    >>> to_flatten = [1, [2,3], [4, [5, [6]], 7], 8]
    >>> assert list(iterflatten(to_flatten)) == [1,2,3,4,5,6,7,8]
    """
    for item in items:
        if isinstance(item, (list, tuple)):
            for itm in iterflatten(item):
                yield itm
        else:
            yield item


def flatten(items):
    """Return flattened list of a list/tuple of lists/tuples::

    >>> assert flatten([1, [2,3], [4, [5, [6]], 7], 8]) == [1,2,3,4,5,6,7,8]
    """
    return list(iterflatten(items))


def iterunique(items):
    """Return iterator to find unique list while preserving the order."""
    seen = []
    for item in items:
        if item not in seen:
            seen.append(item)
            yield item


def unique(items):
    """Return unique list while preserving the order.

    >>> assert unique([1, 2, 3, 1, 2, 3, 4]) == [1, 2, 3, 4]
    """
    return list(iterunique(items))


def mapper_class(relation):
    """Return mapper class given an ORM relation attribute."""
    return relation.property.mapper.class_


def get_mapper_class(model, field):
    """Return mapper class given ORM model and field string."""
    return mapper_class(getattr(model, field))


def get_concrete_value(obj,
                       cls,
                       check_classmethod=False,
                       check_callable=False):
    """Return a 'concrete' form of obj.
    If obj is a declared_attr, it is evaluated on cls.
    If obj is a classmethod and check_classmethod is True,
    it is evaluated on cls.
    If obj is callable and check_callable is True, it is evaluated.
    """
    if isinstance(obj, declared_attr):
        return obj.fget(cls)
    elif check_classmethod and isinstance(obj, classmethod):
        return classmethod_func(obj)(cls)
    elif check_callable and callable(obj):
        return obj()
    else:
        return obj


def merge_declarative_args(cls, global_config_key, local_config_key):
    """Merge declarative args for class `cls`
    identified by `global_config_key` and `local_config_key`
    into a consolidated (tuple, dict).
    """
    configs = [base.__dict__.get(global_config_key)
               for base in reversed(cls.mro())]
    configs.append(cls.__dict__.get(local_config_key))

    args = []
    kargs = {}

    for obj in configs:
        obj = get_concrete_value(obj,
                                 cls,
                                 check_classmethod=True,
                                 check_callable=True)
        if not obj:
            continue

        if isinstance(obj, dict):
            kargs.update(obj)
        elif isinstance(obj[-1], dict):
            # Configuration object has a dict at the end so we assume the
            # passed in format was [item0, item1, ..., dict0].
            args += obj[:-1]
            kargs.update(obj[-1])
        else:
            args += obj

    # Make args unique in case base classes duplicated items.
    args = unique(args)

    return (args, kargs)


def should_set_tablename(cls):
    """Check what values are set by a class and its bases to determine if a
    tablename should be automatically generated.

    The class and its bases are checked in order of precedence: the class
    itself then each base in the order they were given at class definition.

    Abstract classes do not generate a tablename, although they may have set
    or inherited a tablename elsewhere.

    If a class defines a tablename or table, a new one will not be generated.
    Otherwise, if the class defines a primary key, a new name will be
    generated.

    This supports:

    * Joined table inheritance without explicitly naming sub-models.
    * Single table inheritance.
    * Concrete table inheritance
    * Inheriting from mixins or abstract models.

    :param bases: base classes of new class
    :param dct: new class dict
    :return: True if tablename should be set
    """

    dct = cls.__dict__

    if '__tablename__' in dct or '__table__' in dct or '__abstract__' in dct:
        return False

    if AbstractConcreteBase in cls.__bases__:
        return False

    def is_primary_key_column(obj):
        obj = get_concrete_value(obj, cls)
        return isinstance(obj, Column) and obj.primary_key

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",
                                message='Unmanaged access of '
                                        'declarative attribute .*',
                                category=SAWarning)

        for name, value in iteritems(dct):
            if is_primary_key_column(value):
                return True

        if '__mapper_args__' in dct:
            mapper_args = get_concrete_value(dct.get('__mapper_args__', {}),
                                             cls,
                                             check_classmethod=True,
                                             check_callable=True)
            is_concrete = mapper_args.get('concrete', False)
        else:
            mapper_args = get_concrete_value(dct.get('__global_mapper_args__',
                                                     {}),
                                             cls,
                                             check_classmethod=True,
                                             check_callable=True)
            is_concrete = mapper_args.get('concrete', False)
            mapper_args = get_concrete_value(dct.get('__local_mapper_args__',
                                                     {}),
                                             cls,
                                             check_classmethod=True,
                                             check_callable=True)
            is_concrete = mapper_args.get('concrete', is_concrete)

        names_to_ignore = set(dct.keys())
        names_to_ignore.add('query')

        for base in cls.__bases__:
            if (not is_concrete) and (hasattr(base, '__tablename__') or
                                      hasattr(base, '__table__')):
                return False

            for k in base.mro():
                for name, value in iteritems(k.__dict__):
                    if not (name in names_to_ignore or
                            (name.startswith('__') and name.endswith('__'))):
                        if is_primary_key_column(value):
                            return True
                        else:
                            names_to_ignore.add(name)

    return False



"""Project metadata.
"""

__title__ = 'alchy'
__summary__ = ('A SQLAlchemy extension for its declarative ORM that provides '
               'enhancements for model classes, queries, and sessions.')
__url__ = 'https://github.com/dgilland/alchy'

__version__ = '2.2.1'
__install_requires__ = ['SQLAlchemy>=0.9.0', 'pydash>=3.4.3']

__author__ = 'Derrick Gilland'
__email__ = 'dgilland@gmail.com'

__license__ = 'MIT License'



"""Session class that supports multiple database binds.
"""

from sqlalchemy.orm.session import Session as SessionBase


__all__ = [
    'Session'
]


class Session(SessionBase):
    """The default session used by :class:`alchy.manager.Manager`. It extends
    the default session system with bind selection.
    """

    def __init__(self, manager, **options):
        self.manager = manager
        bind = options.pop('bind', manager.engine)
        super(Session, self).__init__(bind=bind,
                                      binds=manager.binds_map,
                                      **options)

    def get_bind(self, mapper=None, clause=None):
        """Return engine bind using mapper info's bind_key if present."""
        # mapper is None if someone tries to just get a connection.
        if mapper is not None:
            info = getattr(mapper.mapped_table, 'info', {})
            bind_key = info.get('bind_key')

            if bind_key is not None:
                return self.manager.get_engine(bind=bind_key)

        return super(Session, self).get_bind(mapper, clause)



"""SQLAlchemy query filter factories usable in
:attr:`alchy.query.QueryModel.__search_filters__`.

These are factory functions that return common filter operations as functions
which are then assigned to the model class' search config attributes. These
functions are syntatic sugar to make it easier to define compatible search
functions. However, due to the fact that a model's query class has to be
defined before the model and given that the model column attributes need to be
defined before using the search factories, there are two ways to use the search
factories on the query class:

1. Define :attr:`alchy.query.QueryModel.__search_filters__` as a property \
that returns the filter dict.
2. Pass in a callable that returns the column.

For example, *without* :mod:`alchy.search` one would define a
:attr:`alchy.query.QueryModel.__search_filters__` similar to::

    class UserQuery(QueryModel):
        __search_filters = {
            'email': lambda value: User.email.like(value)
        }

    class User(Model):
        query_class = UserQuery
        email = Column(types.String(100))

Using :mod:`alchy.search` the above then becomes::

    class UserQuery(QueryModel):
        @property
        def __search_filters__(self):
            return {
                'email': like(User.email)
            }

    class User(Model):
        query_class = UserQuery
        email = Column(types.String(100))

Or if a callable is passed in::

    from alchy import search

    class UserQuery(QueryModel):
        __search_filters__ = {
            'email': like(lambda: User.email)
        }

    class User(Model):
        query_class = UserQuery
        email = Column(types.String(100))

The general naming convention for each comparator is:

- positive comparator: ``<base>`` (e.g. :func:`like`)
- negative comparator: ``not<base>`` (e.g. :func:`notlike`)

The basic call signature for the :mod:`search` functions is::

    # search.<function>(column)
    # e.g.
    search.contains(email)
"""
# pylint: disable=invalid-name

from sqlalchemy import not_


__all__ = [
    'like',
    'notlike',
    'ilike',
    'notilike',
    'startswith',
    'notstartswith',
    'endswith',
    'notendswith',
    'contains',
    'notcontains',
    'icontains',
    'noticontains',
    'in_',
    'notin_',
    'eq',
    'noteq',
    'gt',
    'notgt',
    'ge',
    'notge',
    'lt',
    'notlt',
    'le',
    'notle',
    'any_',
    'notany_',
    'has',
    'nothas',
    'eqenum',
    'noteqenum',
]


class ColumnOperator(object):
    """Base class for column operator based search factories."""
    op = None

    def __init__(self, column):
        self.column = column

    def compare(self, value):
        """Return comparision with value."""
        return getattr(self.column, self.op)(value)

    def __call__(self, value):
        """Makes comparision."""
        if hasattr(self.column, '__call__'):
            self.column = self.column()
        return self.compare(value)


class NegateOperator(ColumnOperator):
    """Negates a ColumnOperator class."""
    def __call__(self, value):
        return not_(super(NegateOperator, self).__call__(value))


class RelationshipOperator(ColumnOperator):
    """Base class for relationship operator based search factories."""
    def __init__(self, column, column_operator):
        self.column = column
        self.column_operator = column_operator

    def compare(self, value):
        """Return comparision with value."""
        return getattr(self.column, self.op)(self.column_operator(value))


class DeclarativeEnumOperator(ColumnOperator):
    """Base class for DeclarativeEnum operator based search factorires."""
    def __init__(self, column, enum_class):
        self.column = column
        self.enum_class = enum_class

    def compare(self, value):
        """Return comparision with value."""
        try:
            return self.column == self.enum_class.from_string(value)
        except ValueError:
            return None


class like(ColumnOperator):
    """Return ``like`` filter function using ORM column field."""
    op = 'like'


class notlike(like, NegateOperator):
    """Return ``not(like)`` filter function using ORM column field."""
    pass


class ilike(ColumnOperator):
    """Return ``ilike`` filter function using ORM column field."""
    op = 'ilike'


class notilike(ilike, NegateOperator):
    """Return ``not(ilike)`` filter function using ORM column field."""
    pass


class startswith(ColumnOperator):
    """Return ``startswith`` filter function using ORM column field."""
    op = 'startswith'


class notstartswith(startswith, NegateOperator):
    """Return ``not(startswith)`` filter function using ORM column field."""
    pass


class endswith(ColumnOperator):
    """Return ``endswith`` filter function using ORM column field."""
    op = 'endswith'


class notendswith(endswith, NegateOperator):
    """Return ``not(endswith)`` filter function using ORM column field."""
    pass


class contains(ColumnOperator):
    """Return ``contains`` filter function using ORM column field."""
    op = 'contains'


class notcontains(contains, NegateOperator):
    """Return ``not(contains)`` filter function using ORM column field."""
    pass


class icontains(ColumnOperator):
    """Return ``icontains`` filter function using ORM column field."""
    def compare(self, value):
        return self.column.ilike('%{0}%'.format(value))


class noticontains(icontains, NegateOperator):
    """Return ``not(icontains)`` filter function using ORM column field."""
    pass


class in_(ColumnOperator):
    """Return ``in_`` filter function using ORM column field."""
    op = 'in_'


class notin_(in_, NegateOperator):
    """Return ``not(in_)`` filter function using ORM column field."""
    pass


class eq(ColumnOperator):
    """Return ``==`` filter function using ORM column field."""
    def compare(self, value):
        return self.column == value


class noteq(eq, NegateOperator):
    """Return ``not(==)`` filter function using ORM column field."""
    pass


class gt(ColumnOperator):
    """Return ``>`` filter function using ORM column field."""
    def compare(self, value):
        return self.column > value


class notgt(gt, NegateOperator):
    """Return ``not(>)`` filter function using ORM column field."""
    pass


class ge(ColumnOperator):
    """Return ``>=`` filter function using ORM column field."""
    def compare(self, value):
        return self.column >= value


class notge(ge, NegateOperator):
    """Return ``not(>=)`` filter function using ORM column field."""
    pass


class lt(ColumnOperator):
    """Return ``<`` filter function using ORM column field."""
    def compare(self, value):
        return self.column < value


class notlt(lt, NegateOperator):
    """Return ``not(<)`` filter function using ORM column field."""
    pass


class le(ColumnOperator):
    """Return ``<=`` filter function using ORM column field."""
    def compare(self, value):
        return self.column <= value


class notle(le, NegateOperator):
    """Return ``not(<=)`` filter function using ORM column field."""
    pass


class any_(RelationshipOperator):
    """Return ``any`` filter function using ORM relationship field."""
    op = 'any'


class notany_(any_, NegateOperator):
    """Return ``not(any)`` filter function using ORM relationship field."""
    pass


class has(RelationshipOperator):
    """Return ``has`` filter function using ORM relationship field."""
    op = 'has'


class nothas(has, NegateOperator):
    """Return ``not(has)`` filter function using ORM relationship field."""
    pass


class eqenum(DeclarativeEnumOperator):
    """Return ``==`` filter function using ORM DeclarativeEnum field."""
    pass


class noteqenum(eqenum, NegateOperator):
    """Return ``not(==)`` filter function using ORM DeclarativeEnum field."""
    pass


class inenum(DeclarativeEnumOperator):
    """Return ``in_`` filter function using ORM DeclarativeEnum field."""
    def compare(self, value):
        """Return comparision with value."""
        if not isinstance(value, (tuple, list)):
            value = [value]

        try:
            return self.column.in_([self.enum_class.from_string(val)
                                    for val in value])
        except ValueError:
            return None


class notinenum(inenum, NegateOperator):
    """Return ``not(in_)`` filter function using ORM DeclarativeEnum field."""
    pass



"""
alchy
=====

A SQLAlchemy extension for its declarative ORM that provides enhancements for
model classes, queries, and sessions.

Project: https://github.com/dgilland/alchy

Documentation: http://alchy.readthedocs.org/
"""

import os
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


meta = {}
exec(read('alchy/__meta__.py'), meta)


class Tox(TestCommand):
    user_options = [
        ('tox-args=', 'a', "Arguments to pass to tox")
    ]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = '-c tox.ini'

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Import here because outside the eggs aren't loaded.
        import tox
        import shlex

        errno = tox.cmdline(args=shlex.split(self.tox_args))
        sys.exit(errno)


setup(
    name=meta['__title__'],
    version=meta['__version__'],
    url=meta['__url__'],
    license=meta['__license__'],
    author=meta['__author__'],
    author_email=meta['__email__'],
    description=meta['__summary__'],
    long_description=read('README.rst'),
    packages=find_packages(exclude=['tests']),
    install_requires=meta['__install_requires__'],
    tests_require=['tox'],
    cmdclass={'test': Tox},
    test_suite='tests',
    keywords='sqlalchemy databases orm declarative',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ]
)



"""Query subclass used by Manager as default session query class.
"""

from math import ceil

from sqlalchemy import orm, and_, or_, inspect
from sqlalchemy.orm.strategy_options import Load
from pydash import py_

from ._compat import iteritems


__all__ = [
    'Query',
    'QueryModel',
    'QueryProperty',
    'Pagination',
    'LoadOption'
]


class Query(orm.Query):
    """Extension of default Query class used in SQLAlchemy session queries.
    """

    #: Default per_page argument for pagination when per_page not specified.
    DEFAULT_PER_PAGE = 50

    @property
    def entities(self):
        """Return list of entity classes present in query."""
        return [e.mapper.class_ for e in self._entities]

    @property
    def join_entities(self):
        """Return list of the joined entity classes present in query."""
        return [e.mapper.class_ for e in self._join_entities]

    @property
    def all_entities(self):
        """Return list of entities + join_entities present in query."""
        return self.entities + self.join_entities

    def _join_eager(self, keys, use_outerjoin, **kargs):
        """Helper method for applying ``join()``/``outerjoin()` with
        ``contains_eager()``.
        """
        alias = kargs.pop('alias', {})
        options = kargs.pop('options', None)

        if not isinstance(alias, dict):
            alias = {keys[0]: alias}

        join_args = [(alias.get(key), key) for key in keys]

        load = orm.contains_eager(keys[0], alias=alias.get(keys[0]))

        for key in keys[1:]:
            load = load.contains_eager(key, alias=alias.get(key))

        if options:
            apply_load_options(load, options)

        join = self.outerjoin if use_outerjoin else self.join

        return join(*join_args).options(load)

    def join_eager(self, *keys, **kargs):
        """Apply ``join`` + ``self.options(contains_eager())``.

        Args:
            keys (mixed): Either string or column references to join
                path(s).

        Keyword Args:
            alias: Join alias or ``dict`` mapping key names to aliases.
            options (list): A list of :class:`LoadOption` to apply to the
                overall load strategy, i.e., each :class:`LoadOption` will be
                chained at the end of the load.
        """
        return self._join_eager(keys, False, **kargs)

    def outerjoin_eager(self, *keys, **kargs):
        """Apply ``outerjoin`` + ``self.options(contains_eager())``.

        Args:
            keys (mixed): Either string keys or column references to join
                path(s).

        Keyword Args:
            alias: Join alias or ``dict`` mapping key names to aliases.
            options (list): A list of :class:`LoadOption` to apply to the
                overall load strategy, i.e., each :class:`LoadOption` will be
                chained at the end of the load.
        """
        return self._join_eager(keys, True, **kargs)

    def _join_load(self, keys, load_strategy, **kargs):
        """Helper method for returning load strategies."""
        options = kargs.pop('options', None)

        load = getattr(orm, load_strategy)(keys[0], **kargs)

        for key in keys[1:]:
            load = getattr(load, load_strategy)(key)

        if options:
            load = apply_load_options(load, options)

        return self.options(load)

    def joinedload(self, *keys, **kargs):
        """Apply ``joinedload()`` to `keys`.

        Args:
            keys (mixed): Either string or column references to join
                path(s).

        Keyword Args:
            options (list): A list of :class:`LoadOption` to apply to the
                overall load strategy, i.e., each :class:`LoadOption` will be
                chained at the end of the load.

        Note:
            Additional keyword args will be passed to initial load creation.
        """
        return self._join_load(keys, 'joinedload', **kargs)

    def immediateload(self, *keys, **kargs):
        """Apply ``immediateload()`` to `keys`.

        Args:
            keys (mixed): Either string or column references to join
                path(s).

        Keyword Args:
            options (list): A list of :class:`LoadOption` to apply to the
                overall load strategy, i.e., each :class:`LoadOption` will be
                chained at the end of the load.

        Note:
            Additional keyword args will be passed to initial load creation.
        """
        return self._join_load(keys, 'immediateload', **kargs)

    def lazyload(self, *keys, **kargs):
        """Apply ``lazyload()`` to `keys`.

        Args:
            keys (mixed): Either string or column references to join
                path(s).

        Keyword Args:
            options (list): A list of :class:`LoadOption` to apply to the
                overall load strategy, i.e., each :class:`LoadOption` will be
                chained at the end of the load.

        Note:
            Additional keyword args will be passed to initial load creation.
        """
        return self._join_load(keys, 'lazyload', **kargs)

    def noload(self, *keys, **kargs):
        """Apply ``noload()`` to `keys`.

        Args:
            keys (mixed): Either string or column references to join
                path(s).

        Keyword Args:
            options (list): A list of :class:`LoadOption` to apply to the
                overall load strategy, i.e., each :class:`LoadOption` will be
                chained at the end of the load.

        Note:
            Additional keyword args will be passed to initial load creation.
        """
        return self._join_load(keys, 'noload', **kargs)

    def subqueryload(self, *keys, **kargs):
        """Apply ``subqueryload()`` to `keys`.

        Args:
            keys (mixed): Either string or column references to join
                path(s).

        Keyword Args:
            options (list): A list of :class:`LoadOption` to apply to the
                overall load strategy, i.e., each :class:`LoadOption` will be
                chained at the end of the load.

        Note:
            Additional keyword args will be passed to initial load creation.
        """
        return self._join_load(keys, 'subqueryload', **kargs)

    def load_only(self, *columns):
        """Apply ``load_only()`` to query."""
        obj, columns = get_load_options(*columns)
        return self.options(obj.load_only(*columns))

    def defer(self, *columns):
        """Apply ``defer()`` to query."""
        load, columns = get_load_options(*columns)
        for column in columns:
            load = load.defer(column)
        return self.options(load)

    def undefer(self, *columns):
        """Apply ``undefer()`` to query."""
        load, columns = get_load_options(*columns)
        for column in columns:
            load = load.undefer(column)
        return self.options(load)

    def undefer_group(self, *names):
        """Apply ``undefer_group()`` to query."""
        obj, names = get_load_options(*names)
        return self.options(obj.undefer_group(names[0]))

    def chain(self):
        """Return pydash chaining instance with items returned by
        :meth:`all`.

        See Also:
            `pydash's <http://pydash.readthedocs.org/>`_ documentation on
            `chaining <http://pydash.readthedocs.org/en/latest/chaining.html>`_
        """
        return py_.chain(self.all())

    def index_by(self, callback=None):
        """Index items returned by :meth:`all` using `callback`."""
        return py_.index_by(self.all(), callback)

    def map(self, callback=None):
        """Map `callback` to each item returned by :meth:`all`."""
        return py_.map(self.all(), callback)

    def reduce(self, callback=None, initial=None):
        """Reduce :meth:`all` using `callback`."""
        return py_.reduce(self.all(), callback, initial)

    def reduce_right(self, callback=None, initial=None):
        """Reduce reversed :meth:`all` using `callback`."""
        return py_.reduce_right(self.all(), callback, initial)

    def pluck(self, column):
        """Pluck `column` attribute values from :meth:`all` results and
        return as list.
        """
        return py_.pluck(self.all(), column)

    def page(self, page=1, per_page=None):
        """Return query with limit and offset applied for page."""
        if per_page is None:
            per_page = self.DEFAULT_PER_PAGE

        return self.limit(per_page).offset((page - 1) * per_page)

    def paginate(self, page=1, per_page=None, error_out=True):
        """Return :class:`Pagination` instance using already defined query
        parameters.
        """
        if error_out and page < 1:
            raise IndexError

        if per_page is None:
            per_page = self.DEFAULT_PER_PAGE

        items = self.page(page, per_page).all()

        if not items and page != 1 and error_out:
            raise IndexError

        # No need to count if we're on the first page and there are fewer items
        # than we expected.
        if page == 1 and len(items) < per_page:
            total = len(items)
        else:
            total = self.order_by(None).count()

        return Pagination(self, page, per_page, total, items)


class QueryModel(Query):
    """Class used for default query property class for ``mymanager.query``,
    ``mymanager.session.query``, and ``MyModel.query``. Can be used in other
    libraries/implementations when creating a session::

        from sqlalchemy import orm

        from alchy import QueryModel
        # or if not using as query property
        # from alchy import Query


        session = orm.scoped_session(orm.sessionmaker())
        session.configure(query_cls=QueryModel)

    **NOTE:** If you don't plan to use the query class as a query property,
    then you can use the :class:`Query` class instead since it won't include
    features that only work within a query property context.

    Attributes:

        __search_filters__: All available search filter functions indexed by a
            canonical name which will be referenced in advanced/simple search.
            All filter functions should take a single value and return an
            SQLAlchemy filter expression, i.e.,
            ``{key: lambda value: Model.column_name.contains(value)}``

        __advanced_search__: Advanced search models search by named parameters.
            Generally found on advanced search forms where each field maps to a
            specific database field that will be queried against. If defined as
            a list, each item should be a key from :attr:`__search_filters__`.
            The matching :attr:`__search_filters__` function will be used in
            the query. If defined as a dict, it should have the same format as
            :attr:`__search_filters__`.

        __simple_search__: Simple search models search by phrase (like Google
            search). Defined like :attr:`__advanced_search__`.

        __order_by__: Default order-by to use when
            :attr:`alchy.model.ModelBase.query` used.
    """

    __search_filters__ = {}
    __advanced_search__ = []
    __simple_search__ = []
    __order_by__ = None

    @property
    def Model(self):
        """Return primary entity model class."""
        return self.entities[0]

    def get_search_filters(self, keys):
        """Return :attr:`__search_filters__` filtered by keys."""
        if isinstance(keys, dict):
            return keys
        else:
            return dict([(key, self.__search_filters__[key]) for key in keys])

    def advanced_filter(self, search_dict=None):
        """Return the compiled advanced search filter mapped to `search_dict`.
        """
        if search_dict is None:  # pragma: no cover
            search_dict = {}

        filter_funcs = self.get_search_filters(self.__advanced_search__)
        term_filters = [filter_funcs[key](value)
                        for key, value in iteritems(search_dict)
                        if key in filter_funcs]

        # All filters should match for an advanced search.
        return and_(*term_filters)

    def simple_filter(self, search_terms=None):
        """Return the compiled simple search filter mapped to `search_terms`.
        """
        if search_terms is None:  # pragma: no cover
            search_terms = []

        filter_funcs = self.get_search_filters(self.__simple_search__)

        # Only support AND'ing search terms together. Apply each simple search
        # filter to each search term and group them together.
        term_filters = [[func(term) for func in filter_funcs.values()]
                        for term in search_terms]

        # Each item in term_filters is a list of filters applied to one of
        # the search terms contained in search_string. We need at least one
        # simple filter to match for each term. We need all search terms to
        # have at least simple filter match.
        return and_(*[or_(*filters) for filters in term_filters])

    def search(self, search_string=None, search_dict=None, **search_options):
        """Perform combination of simple/advanced searching with optional
        limit/offset support.
        """
        search_options.setdefault('limit', None)
        search_options.setdefault('offset', None)
        search_options.setdefault('order_by', self.__order_by__)

        query = self

        # Apply search filtering and pagination to Model's primary keys so we
        # can use the query as a subquery. In order to properly handle
        # pagination, we can use a subquery so that the outer level joins
        # won't cause records to be excluded when they include *-to-many
        # relationships. For example, if we were returning a query of user +
        # user keywords (one-to-many), then for something like the first 25
        # users, we may actually have more than that many records since we're
        # joining on many records from the user keywords table.
        original = (self.lazyload('*')
                    .load_only(*self.Model.primary_attrs())
                    .distinct())

        # Use the original query so that we preserve joins and where
        # statements.
        model_query = original

        if self.whereclause is not None:
            # If our base query contains a whereclause, then we need to
            # compelete the "transfer" of the base query's where statements to
            # model_query by wiping out the base query's criterion. i.e. We
            # only want to maintain selects and froms in the base query and
            # keep wheres in the model_query.

            # Call a generative query method that won't modify its state. This
            # is basically a no-op used to copy the query object and modify it
            # below. NOTE: There may be a better way to do this.
            query = query.filter()

            # Remove existing filters since they were transferred to the
            # model_query. This may seem kind of hacky but I don't know of a
            # better way to nullify the query object's where clause.
            query._criterion = None

        if search_string is not None:
            model_query = model_query.filter(
                self.simple_filter(search_string.split()))

        if search_dict is not None:
            model_query = model_query.filter(
                self.advanced_filter(search_dict))

        if search_options['order_by'] is not None:
            if not isinstance(search_options['order_by'], (list, tuple)):
                search_options['order_by'] = [search_options['order_by']]
            model_query = model_query.order_by(*search_options['order_by'])

        if search_options['limit'] is not None:
            model_query = model_query.limit(search_options['limit'])

        if search_options['offset'] is not None:
            model_query = model_query.offset(search_options['offset'])

        if model_query != original:
            subquery = model_query.subquery()
            query = query.join(
                subquery, join_subquery_on_columns(subquery,
                                                   self.Model.primary_keys()))

        return query


class QueryProperty(object):
    """Query property accessor which gives a model access to query capabilities
    via :attr:`alchy.model.ModelBase.query` which is equivalent to
    ``session.query(Model)``.
    """
    def __init__(self, session):
        self.session = session

    def __get__(self, model, Model):
        mapper = orm.class_mapper(Model)
        if mapper:
            if not getattr(Model, 'query_class', None):
                Model.query_class = QueryModel

            query_property = Model.query_class(mapper, session=self.session())

            return query_property


##
# Pagination class and usage adapated from Flask-SQLAlchemy:
# https://github.com/mitsuhiko/flask-sqlalchemy
##


class Pagination(object):
    """Internal helper class returned by :meth:`Query.paginate`. You can also
    construct it from any other SQLAlchemy query object if you are working with
    other libraries. Additionally it is possible to pass ``None`` as query
    object in which case the `prev` and `next` will no longer work.
    """

    def __init__(self, query, page, per_page, total, items):
        #: The query object that was used to create this pagination object.
        self.query = query

        #: The current page number (1 indexed).
        self.page = page

        #: The number of items to be displayed on a page.
        self.per_page = per_page

        #: The total number of items matching the query.
        self.total = total

        #: The items for the current page.
        self.items = items

        if self.per_page == 0:
            self.pages = 0
        else:
            #: The total number of pages.
            self.pages = int(ceil(self.total / float(self.per_page)))

        #: Number of the previous page.
        self.prev_num = self.page - 1

        #: True if a previous page exists.
        self.has_prev = self.page > 1

        #: Number of the next page.
        self.next_num = self.page + 1

        #: True if a next page exists.
        self.has_next = self.page < self.pages

    def prev(self, error_out=False):
        """Returns a :class:`Pagination` object for the previous page."""
        assert self.query is not None, \
            'a query object is required for this method to work'
        return self.query.paginate(self.page - 1, self.per_page, error_out)

    def next(self, error_out=False):
        """Returns a :class:`Pagination` object for the next page."""
        assert self.query is not None, \
            'a query object is required for this method to work'
        return self.query.paginate(self.page + 1, self.per_page, error_out)


class LoadOption(object):
    """Chained load option to apply to a load strategy when calling
    :class:`Query` load methods.

    Example usage: ::

        qry = (db.session.query(Product)
               .join_eager('category',
                           options=[LoadOption('noload', 'images')]))

    This would result in the ``noload`` option being chained to the eager
    option for ``Product.category`` and is equilvalent to: ::

        qry = (db.session.query(Product)
               .join('category')
               .options(contains_eager('category').noload('images')))
    """
    def __init__(self, strategy, *args, **kargs):
        self.strategy = strategy
        self.args = args
        self.kargs = kargs


def get_load_options(*columns):
    """Helper method that attempts to extract a sqlalchemy object from
    `columns[0]` and return remaining columns to apply to a query load method.
    """
    model_inspect = inspect(columns[0], raiseerr=False)

    # return an obj which has loading API
    if model_inspect and model_inspect.is_mapper:
        obj = Load(columns[0])
        columns = columns[1:]
    elif isinstance(columns[0], Load):
        obj = columns[0]
        columns = columns[1:]
    else:
        obj = orm

    return (obj, columns)


def apply_load_options(load, options):
    """Apply load `options` to base `load` object.
    """
    for load_option in options:
        load = getattr(load, load_option.strategy)(*load_option.args,
                                                   **load_option.kargs)

    return load


def base_columns_from_subquery(subquery):
    """Return non-aliased, base columns from subquery."""
    # base_columns is a set so we need to cast to list.
    return [(column, list(column.base_columns))
            for column in subquery.c.values()]


def join_subquery_on_columns(subquery, columns):
    """Return join-on condition which maps subquery's columns to columns."""
    subquery_base_columns = base_columns_from_subquery(subquery)

    join_on = []
    for subquery_column, base_columns in subquery_base_columns:
        # Don't support joining to subquery column with more than 1 base
        # column.
        if len(base_columns) == 1 and base_columns[0] in columns:
            join_on.append(subquery_column == base_columns[0])

    if join_on:
        return and_(*join_on)
    else:  # pragma: no cover
        return None
